"use strict";
var ts = require("typescript");
var utils_1 = require("./../../utils");
var TsTypeChecker = (function () {
    function TsTypeChecker(typeChecker) {
        this.typeChecker = typeChecker;
    }
    TsTypeChecker.prototype.typeToString = function (type, node) {
        var formatFlags = (ts.TypeFormatFlags.UseTypeOfFunction | ts.TypeFormatFlags.NoTruncation | ts.TypeFormatFlags.UseFullyQualifiedType |
            ts.TypeFormatFlags.WriteTypeArgumentsOfSignature);
        if (node.kind === ts.SyntaxKind.TypeAliasDeclaration) {
            formatFlags |= ts.TypeFormatFlags.InTypeAlias;
        }
        return this.typeChecker.typeToString(type, node, formatFlags);
    };
    TsTypeChecker.prototype.getSyntaxKindAsString = function (kind) {
        return ts.SyntaxKind[kind];
    };
    TsTypeChecker.prototype.getConstantValue = function (node) {
        return this.typeChecker.getConstantValue(node);
    };
    TsTypeChecker.prototype.getReturnTypeOfSignature = function (signature) {
        return this.typeChecker.getReturnTypeOfSignature(signature);
    };
    TsTypeChecker.prototype.getTypeAtLocation = function (node) {
        return this.typeChecker.getTypeAtLocation(node);
    };
    TsTypeChecker.prototype.getDeclarationFromSymbol = function (symbol) {
        if (symbol.valueDeclaration != null) {
            return symbol.valueDeclaration;
        }
        else {
            var declarations = symbol.getDeclarations();
            /* istanbul ignore next */
            if (declarations == null) {
                throw new Error("Declaration should not be null for symbol: " + symbol.name);
            }
            else if (declarations.length > 1) {
                utils_1.Logger.warn("Not implemented. Symbol has more than one declaration: " + symbol.name);
            }
            else if (declarations.length === 0) {
                throw new Error("Declaration length should not be 0 for symbol: " + symbol.name);
            }
            return declarations[0];
        }
    };
    TsTypeChecker.prototype.getAliasedSymbol = function (symbol) {
        if (this.symbolHasFlag(symbol, ts.SymbolFlags.Alias)) {
            return this.typeChecker.getAliasedSymbol(symbol);
        }
        else {
            return null;
        }
    };
    TsTypeChecker.prototype.getLocalSymbolFromNode = function (node) {
        return node["localSymbol"];
    };
    TsTypeChecker.prototype.getLocalSymbolsFromNode = function (node) {
        var locals = node.locals;
        /* istanbul ignore if */
        if (locals == null) {
            return [];
        }
        else {
            return Object.keys(locals).map(function (key) {
                var symbol = locals[key];
                var exportSymbol = symbol.exportSymbol;
                return exportSymbol || symbol;
            });
        }
    };
    TsTypeChecker.prototype.getReturnTypeOfNode = function (node) {
        return this.typeChecker.getReturnTypeOfSignature(this.getSignatureFromNode(node));
    };
    TsTypeChecker.prototype.getSignatureFromNode = function (node) {
        return this.typeChecker.getSignatureFromDeclaration(node);
    };
    TsTypeChecker.prototype.getSymbolAtLocation = function (node) {
        return node.symbol || this.typeChecker.getSymbolAtLocation(node);
    };
    TsTypeChecker.prototype.getExportsOfModule = function (symbol) {
        return this.typeChecker.getExportsOfModule(symbol);
    };
    /* istanbul ignore next */
    TsTypeChecker.prototype.getInternalTypeCheckerForTesting = function () {
        // get the type checker for quick testing purposes.
        // this should never be used in the code!
        return this.typeChecker;
    };
    TsTypeChecker.prototype.isSymbolNamedExportOfFile = function (symbol, file) {
        var fileSymbol = this.getSymbolAtLocation(file);
        // when a file doesn't have exports the symbol will be null
        return fileSymbol != null && fileSymbol.exports[symbol.name] != null && symbol.name !== "default";
    };
    TsTypeChecker.prototype.isSymbolDefaultExportOfFile = function (symbol, file) {
        var fileSymbol = this.getSymbolAtLocation(file);
        var isDefaultExportOfFile = false;
        if (fileSymbol != null) {
            var defaultExport = fileSymbol.exports["default"];
            if (defaultExport != null) {
                isDefaultExportOfFile = defaultExport === symbol || this.getAliasedSymbol(defaultExport) === symbol;
            }
        }
        return isDefaultExportOfFile;
    };
    TsTypeChecker.prototype.symbolHasFlag = function (symbol, flag) {
        return (symbol.getFlags() & flag) !== 0;
    };
    return TsTypeChecker;
}());
exports.TsTypeChecker = TsTypeChecker;

//# sourceMappingURL=TsTypeChecker.js.map
