"use strict";
var assert = require("assert");
var path = require("path");
var definitions_1 = require("./../../definitions");
var main_1 = require("./../../main");
var createFunctions_1 = require("./../../createFunctions");
describe("GlobalDefinition", function () {
    describe("#renameDefinitionAs()", function () {
        describe("passing in a definition that doesn't exist in the global definition", function () {
            it("should not error", function () {
                var originalCode = "var myString: string;\n";
                var file = main_1.getInfoFromString(originalCode);
                var globalDef = new definitions_1.GlobalDefinition();
                globalDef.files.push(file);
                globalDef.renameDefinitionAs(createFunctions_1.createVariable({ name: "myOtherVar" }), "MyNewName");
                assert.equal(file.write(), originalCode);
            });
        });
        describe("renaming for internal modules", function () {
            var globalDef = new definitions_1.GlobalDefinition();
            globalDef.addFile({
                namespaces: [{
                        name: "MyNamespace",
                        variables: [{
                                name: "c",
                                defaultExpression: "MyClass"
                            }],
                        classes: [{
                                name: "MyClass"
                            }]
                    }],
                variables: [{
                        name: "d",
                        defaultExpression: "MyNamespace.MyClass"
                    }]
            });
            globalDef.addFile({
                variables: [{
                        name: "a",
                        type: "typeof MyNamespace",
                        defaultExpression: "MyNamespace"
                    }]
            });
            globalDef.addFile({
                variables: [{
                        name: "b",
                        defaultExpression: "a.MyClass"
                    }]
            });
            globalDef.renameDefinitionAs(globalDef.files[0].namespaces[0].classes[0], "MyNewName");
            it("should have the correct code in the main file", function () {
                var expectedCode = "namespace MyNamespace {\n    class MyNewName {\n    }\n\n    let c = MyNewName;\n}\n\nlet d = MyNamespace.MyNewName;\n";
                assert.equal(globalDef.files[0].write(), expectedCode);
            });
            it("should have the correct code in the other file", function () {
                var expectedCode = "let b = a.MyNewName;\n";
                assert.equal(globalDef.files[2].write(), expectedCode);
            });
        });
        describe("renaming without imports", function () {
            var code = "namespace MyNamespace {\n    interface MyInterface extends MyClass {\n        prop: MyClass;\n\n        method2(m: MyClass): void;\n    }\n}\n\nclass MyOtherClass<T> {\n}\n\nclass MyClass extends MyOtherClass<MyClass> {\n    property = new MyClass();\n\n    method(p = new MyClass()): p is MyClass {\n    }\n}\n\nlet myVar: MyClass & MyOtherClass<MyClass>;\n";
            var file = main_1.getInfoFromString(code);
            var globalDef = new definitions_1.GlobalDefinition();
            globalDef.files.push(file);
            globalDef.renameDefinitionAs(file.getClass("MyClass"), "MyNewName");
            it("should rename the definition", function () {
                assert.equal(file.write(), code.replace(/MyClass/g, "MyNewName"));
            });
        });
        describe("renaming with assignment to variable", function () {
            var code = "namespace MyNamespace {\n    var f = MyNamespace;\n    var g = f.MyInnerNamespace.MyClass;\n    var h = MyInnerNamespace;\n    var i = h.MyClass;\n\n    export namespace MyInnerNamespace {\n        var j = MyInnerNamespace;\n        var k = h.MyClass;\n\n        export class MyClass {\n        }\n    }\n\n    namespace MyOtherInnerNamespace {\n        var l = h.MyClass;\n    }\n}\n\nvar a = MyNamespace;\nvar b = a.MyInnerNamespace;\nvar c = MyNamespace.MyInnerNamespace;\nvar d = b.MyClass;\nvar e = c.MyClass;\n";
            var file = main_1.getInfoFromString(code);
            var globalDef = new definitions_1.GlobalDefinition();
            globalDef.files.push(file);
            globalDef.renameDefinitionAs(file.getNamespace("MyNamespace").getNamespace("MyInnerNamespace").getClass("MyClass"), "MyNewName");
            it("should rename the definition", function () {
                assert.equal(file.write(), code.replace(/MyClass/g, "MyNewName"));
            });
        });
        describe("renaming with imports and no namespace", function () {
            function getGlobalDef(_a) {
                var _b = _a.aliasName, aliasName = _b === void 0 ? "" : _b, _c = _a.isDefaultImport, isDefaultImport = _c === void 0 ? false : _c;
                var globalDef = new definitions_1.GlobalDefinition();
                globalDef.addFile({
                    fileName: "C:\\MyClass.ts",
                    classes: [{
                            name: "MyClass",
                            isNamedExportOfFile: !isDefaultImport,
                            isDefaultExportOfFile: isDefaultImport
                        }, {
                            name: "MyOtherClass",
                            isNamedExportOfFile: true
                        }],
                    variables: [{
                            name: "myVar",
                            type: "MyClass"
                        }],
                    defaultExportExpression: isDefaultImport ? "MyClass" : undefined
                });
                globalDef.addFile({
                    fileName: "C:\\Main.ts",
                    variables: [{
                            name: "myVar",
                            type: aliasName || "MyClass"
                        }, {
                            name: "myVar2",
                            type: "MyClass"
                        }],
                    imports: [{
                            defaultImportName: "dummy",
                            moduleSpecifier: "./dummyImport"
                        }]
                });
                globalDef.addDefinitionAsImportToFile({
                    definition: globalDef.files[0].classes[0],
                    file: globalDef.files[1],
                    alias: aliasName
                });
                globalDef.addDefinitionAsImportToFile({
                    definition: globalDef.files[0].classes[1],
                    file: globalDef.files[1]
                });
                return globalDef;
            }
            describe("renaming with imports and no alias", function () {
                var expectedCode = "import dummy from \"./dummyImport\";\nimport {MyNewName} from \"./MyClass\";\nimport {MyOtherClass} from \"./MyClass\";\n\nlet myVar: MyNewName;\nlet myVar2: MyNewName;\n";
                var globalDef = getGlobalDef({});
                globalDef.renameDefinitionAs(globalDef.files[0].classes[0], "MyNewName");
                it("should write the file, change the imports, and change the reference in the file", function () {
                    assert.equal(globalDef.files[1].write(), expectedCode);
                });
                it("should rename the definition", function () {
                    assert.equal(globalDef.files[0].classes[0].name, "MyNewName");
                });
            });
            describe("renaming with imports and an alias", function () {
                var expectedCode = "import dummy from \"./dummyImport\";\nimport {MyNewName as MyAlias} from \"./MyClass\";\nimport {MyOtherClass} from \"./MyClass\";\n\nlet myVar: MyAlias;\nlet myVar2: MyClass;\n";
                it("should write the file and only change the import", function () {
                    var globalDef = getGlobalDef({ aliasName: "MyAlias" });
                    globalDef.renameDefinitionAs(globalDef.files[0].classes[0], "MyNewName");
                    assert.equal(globalDef.files[1].write(), expectedCode);
                });
            });
            describe("renaming with default import", function () {
                var globalDef = getGlobalDef({ aliasName: "MyClass", isDefaultImport: true });
                globalDef.renameDefinitionAs(globalDef.files[0].classes[0], "MyNewName");
                it("should rename in the file the default import is contained in", function () {
                    var expectedCode = "class MyNewName {\n}\n\nexport class MyOtherClass {\n}\n\nlet myVar: MyNewName;\n\nexport default MyNewName;\n";
                    assert.equal(globalDef.files[0].write(), expectedCode);
                });
                it("should not rename anything in the other file", function () {
                    var expectedCode = "import dummy from \"./dummyImport\";\nimport {default as MyClass} from \"./MyClass\";\nimport {MyOtherClass} from \"./MyClass\";\n\nlet myVar: MyClass;\nlet myVar2: MyClass;\n";
                    assert.equal(globalDef.files[1].write(), expectedCode);
                });
            });
        });
        describe("renaming with imports and namespace", function () {
            var globalDef = new definitions_1.GlobalDefinition();
            globalDef.addFile({
                fileName: "C:/MyNamespace.ts",
                namespaces: [{
                        name: "MyNamespace",
                        isNamedExportOfFile: true,
                        classes: [{
                                name: "MyClass",
                                isExported: true
                            }]
                    }, {
                        name: "MyDefaultNamespace",
                        isDefaultExportOfFile: true,
                        classes: [{
                                name: "MyClass",
                                isExported: true,
                                extendsTypes: ["MyNamespace.MyClass"]
                            }]
                    }],
                defaultExportExpression: "MyDefaultNamespace"
            });
            globalDef.addFile({
                fileName: "C:/main.ts",
                imports: [{
                        moduleSpecifier: "./MyNamespace",
                        namedImports: [{
                                name: "MyNamespace"
                            }, {
                                name: "MyNamespace",
                                alias: "MyNamespaceAlias"
                            }, {
                                name: "default",
                                alias: "MyDefaultNamedNamespace"
                            }]
                    }, {
                        moduleSpecifier: "./MyNamespace",
                        starImportName: "MyNamespaceModule"
                    }, {
                        moduleSpecifier: "./MyNamespace",
                        defaultImportName: "MyDefaultImportNamespace"
                    }],
                variables: [{
                        name: "myVar",
                        type: "typeof MyNamespace.MyClass"
                    }, {
                        name: "myNamespaceAliasVar",
                        type: "typeof MyNamespaceAlias.MyClass"
                    }, {
                        name: "myDefaultNamedNamespaceVar",
                        type: "typeof MyDefaultNamedNamespace.MyClass"
                    }, {
                        name: "myVar2",
                        type: "typeof MyNamespaceModule.MyNamespace.MyClass"
                    }, {
                        name: "myNamespaceVar",
                        type: "typeof MyNamespace",
                        defaultExpression: "MyNamespace"
                    }, {
                        name: "myVar3",
                        type: "typeof MyNamespace.MyClass",
                        defaultExpression: "myNamespaceVar.MyClass"
                    }, {
                        name: "myDefaultImportVar",
                        type: "typeof MyDefaultImportNamespace.MyClass"
                    }]
            });
            describe("renaming the classes", function () {
                globalDef.renameDefinitionAs(globalDef.files[0].namespaces[0].classes[0], "MyNewName");
                globalDef.renameDefinitionAs(globalDef.files[0].namespaces[1].classes[0], "MyOtherNewName");
                it("should rename the class in the main file", function () {
                    var expectedCode = "export namespace MyNamespace {\n    export class MyNewName {\n    }\n}\n\nnamespace MyDefaultNamespace {\n    export class MyOtherNewName extends MyNamespace.MyNewName {\n    }\n}\n\nexport default MyDefaultNamespace;\n";
                    assert.equal(globalDef.files[0].write(), expectedCode);
                });
                it("should rename the class in the other file", function () {
                    var expectedCode = "import {MyNamespace, MyNamespace as MyNamespaceAlias, default as MyDefaultNamedNamespace} from \"./MyNamespace\";\nimport * as MyNamespaceModule from \"./MyNamespace\";\nimport MyDefaultImportNamespace from \"./MyNamespace\";\n\nlet myVar: typeof MyNamespace.MyNewName;\nlet myNamespaceAliasVar: typeof MyNamespaceAlias.MyNewName;\nlet myDefaultNamedNamespaceVar: typeof MyDefaultNamedNamespace.MyOtherNewName;\nlet myVar2: typeof MyNamespaceModule.MyNamespace.MyNewName;\nlet myNamespaceVar = MyNamespace;\nlet myVar3 = myNamespaceVar.MyNewName;\nlet myDefaultImportVar: typeof MyDefaultImportNamespace.MyOtherNewName;\n";
                    assert.equal(globalDef.files[1].write(), expectedCode);
                });
            });
        });
        describe("renaming re-exported definitions", function () {
            var globalDef = new definitions_1.GlobalDefinition();
            globalDef.addFile({
                fileName: "C:/MyClass.ts",
                classes: [{
                        name: "MyClass",
                        isNamedExportOfFile: true
                    }],
                namespaces: [{
                        name: "MyNamespace",
                        classes: [{
                                name: "MyInnerClass",
                                isExported: true
                            }],
                        isDefaultExportOfFile: true
                    }],
                defaultExportExpression: "MyNamespace"
            });
            globalDef.addFile({
                fileName: "C:/classes.ts",
                reExports: [{
                        moduleSpecifier: "./MyClass"
                    }]
            });
            globalDef.addFile({
                fileName: "C:/namedClasses.ts",
                reExports: [{
                        moduleSpecifier: "./MyClass",
                        namedExports: [{
                                name: "MyClass"
                            }, {
                                name: "MyClass",
                                alias: "MyClassAlias"
                            }, {
                                name: "default",
                                alias: "MyNamespace"
                            }]
                    }]
            });
            globalDef.addFile({
                fileName: "C:/main.ts",
                imports: [{
                        moduleSpecifier: "./classes",
                        namedImports: [{
                                name: "MyClass"
                            }, {
                                name: "MyClass",
                                alias: "MyClassAlias"
                            }]
                    }, {
                        moduleSpecifier: "./namedClasses",
                        namedImports: [{
                                name: "MyClass",
                                alias: "MyClassAlias2"
                            }, {
                                name: "MyNamespace",
                                alias: "MyAliasNamespace"
                            }]
                    }],
                variables: [{
                        name: "a",
                        type: "MyClass"
                    }, {
                        name: "b",
                        type: "MyClassAlias"
                    }, {
                        name: "c",
                        type: "MyAliasNamespace.MyInnerClass"
                    }]
            });
            var file = globalDef.getFile("MyClass.ts");
            globalDef.renameDefinitionAs(file.classes[0], "MyNewName");
            globalDef.renameDefinitionAs(file.namespaces[0].classes[0], "MyNewInnerClass");
            it("should rename the re-exports appropriately", function () {
                var expectedCode = "export {MyNewName, MyNewName as MyClassAlias, default as MyNamespace} from \"./MyClass\";\n";
                assert.equal(globalDef.getFile("namedClasses.ts").write(), expectedCode);
            });
            it("should rename the imported re-exported definition in the main file", function () {
                var expectedCode = "import {MyNewName, MyNewName as MyClassAlias} from \"./classes\";\nimport {MyNewName as MyClassAlias2, MyNamespace as MyAliasNamespace} from \"./namedClasses\";\n\nlet a: MyNewName;\nlet b: MyClassAlias;\nlet c: MyAliasNamespace.MyNewInnerClass;\n";
                assert.equal(globalDef.getFile("main.ts").write(), expectedCode);
            });
        });
        describe("re-naming default exports", function () {
            var globalDef = main_1.getInfoFromFiles([path.join(__dirname, "./../../../src/tests/testFiles/reExportRename/main.ts")]);
            globalDef.renameDefinitionAs(globalDef.getFile("MyNamespace.ts").namespaces[0].classes[0], "MyNewClass");
            it("should rename in the main file correctly", function () {
                var expectedCode = "import * as StarImport from \"./reExports\";\nimport {MyNamespace, MyNamespace as MyNamespaceAlias} from \"./reExports\";\n\nlet a = StarImport.MyNamespace.MyNewClass;\nlet b = MyNamespace.MyNewClass;\nlet c = MyNamespaceAlias.MyNewClass;\n";
                assert.equal(globalDef.getFile("main.ts").write(), expectedCode);
            });
        });
    });
});

//# sourceMappingURL=globalDefinitionRenameDefinitionAsTests.js.map
