"use strict";
var assert = require("assert");
var createFunctions_1 = require("./../../createFunctions");
var definitions_1 = require("./../../definitions");
describe("GlobalDefinition", function () {
    describe("#addDefinitionAsImportToFile()", function () {
        var def = new definitions_1.GlobalDefinition();
        def.addFile({
            fileName: "/file1.ts",
            classes: [{
                    name: "MyClass",
                    isNamedExportOfFile: true
                }, {
                    name: "MyOtherClass",
                    isDefaultExportOfFile: true
                }, {
                    name: "MyNonExportedClass"
                }],
            namespaces: [{
                    name: "MyDefaultNamespace",
                    isDefaultExportOfFile: true,
                    classes: [{
                            name: "MyClassInNamespace",
                            isExported: true
                        }]
                }]
        });
        def.addFile({
            fileName: "/file2.ts"
        });
        var file1 = def.getFile("file1.ts");
        var file2 = def.getFile("file2.ts");
        it("should throw an error when specifying a definition that is not exported", function () {
            assert.throws(function () { return def.addDefinitionAsImportToFile({ definition: file1.getClass("MyNonExportedClass"), file: file2 }); });
        });
        it("should throw an error when specifying a definition that does not exist in the global definition", function () {
            assert.throws(function () { return def.addDefinitionAsImportToFile({ definition: createFunctions_1.createVariable({ name: "myVar", isNamedExportOfFile: true }), file: file2 }); });
        });
        describe("adding an exported definition as an import to another file", function () {
            def.addDefinitionAsImportToFile({ definition: file1.getClass("MyClass"), file: file2 });
            var importDef = file2.imports[0];
            it("should have a name of MyClass", function () {
                assert.equal(importDef.namedImports[0].name, "MyClass");
            });
            it("should have a definition of MyClass", function () {
                assert.equal(importDef.namedImports[0].definitions[0].name, "MyClass");
            });
            it("should have a module specifier of ./file1", function () {
                assert.equal(importDef.moduleSpecifier, "./file1");
            });
        });
        describe("adding an exported definition as an import inside a namespace", function () {
            def.addDefinitionAsImportToFile({ definition: file1.getNamespace("MyDefaultNamespace").getClass("MyClassInNamespace"), file: file2 });
            var importDef = file2.imports[1];
            it("should have a definition of MyDefaultNamespace", function () {
                assert.equal(importDef.namedImports[0].definitions[0].name, "MyDefaultNamespace");
            });
            it("should have a name of default", function () {
                assert.equal(importDef.namedImports[0].name, "default");
            });
            it("should have an alias of MyDefaultNamespace", function () {
                assert.equal(importDef.namedImports[0].alias, "MyDefaultNamespace");
            });
            it("should have a module specifier of ./file1", function () {
                assert.equal(importDef.moduleSpecifier, "./file1");
            });
        });
        describe("adding an exported definition as an import to another file with an alias", function () {
            def.addDefinitionAsImportToFile({ definition: file1.getClass("MyClass"), file: file2, alias: "myAlias" });
            var importDef = file2.imports[2];
            it("should have an alias of myAlias", function () {
                assert.equal(importDef.namedImports[0].alias, "myAlias");
            });
        });
    });
    describe("#addFile()", function () {
        var def = new definitions_1.GlobalDefinition();
        var returnedDef = def.addFile({ fileName: "test.ts" });
        def.addFile({ fileName: "test2.ts" });
        it("the returned definition should be in the array", function () {
            assert.equal(returnedDef, def.files[0]);
        });
        // good enough to just check the number of files is correct...
        // this is tested more thoroughly in the createFile tests
        it("should have the right number of files", function () {
            assert.equal(def.files.length, 2);
        });
    });
    describe("#getFile()", function () {
        var def = new definitions_1.GlobalDefinition();
        def.addFile({ fileName: "test.ts" });
        def.addFile({ fileName: "test2.ts" });
        it("should get the correct file when specifying a file name", function () {
            assert.equal(def.getFile("test2.ts"), def.files[1]);
        });
        it("should get the correct file when specifying a function", function () {
            assert.equal(def.getFile(function (f) { return f.fileName === "test2.ts"; }), def.files[1]);
        });
    });
    describe("#getFileOfDefinition()", function () {
        // no need to test anything extravagant in here because this method calls getNamespacesToDefinition()
        var def = new definitions_1.GlobalDefinition();
        def.addFile({
            fileName: "dummy.ts"
        });
        def.addFile({
            namespaces: [{
                    name: "n",
                    variables: [{
                            name: "v"
                        }]
                }]
        });
        it("should get the correct file when specifying something in a file", function () {
            assert.equal(def.getFileOfDefinition(def.files[1].namespaces[0].variables[0]), def.files[1]);
        });
        it("should return null when specifying something not in a file", function () {
            assert.equal(def.getFileOfDefinition(createFunctions_1.createVariable({ name: "v" })), null);
        });
    });
    describe("#getNamespacesToDefinition()", function () {
        var def = new definitions_1.GlobalDefinition();
        def.addFile({
            fileName: "dummy.ts"
        });
        def.addFile({
            namespaces: [{
                    name: "n1"
                }, {
                    name: "n2",
                    namespaces: [{
                            name: "n3",
                            variables: [{ name: "v1" }, { name: "v2" }]
                        }]
                }]
        });
        def.addFile({
            variables: [{ name: "v" }]
        });
        describe("getting the file and namespaces to a variable in a file", function () {
            var result = def.getFileAndNamespacesToDefinition(def.files[2].variables[0]);
            it("should have the correct file", function () {
                assert.equal(result.file, def.files[2]);
            });
            it("the array should have the correct namespace length", function () {
                assert.equal(result.namespaces.length, 0);
            });
        });
        describe("getting the file and namespaces to a variable in a file within a namespace", function () {
            var result = def.getFileAndNamespacesToDefinition(def.files[1].namespaces[1].namespaces[0].variables[1]);
            it("should have the correct file", function () {
                assert.equal(result.file, def.files[1]);
            });
            it("the array should have the correct namespace length", function () {
                assert.equal(result.namespaces.length, 2);
            });
            it("should have the first namespace as the first item in the array", function () {
                assert.equal(result.namespaces[0], def.files[1].namespaces[1]);
            });
            it("should have the second namespace as the second item in the array", function () {
                assert.equal(result.namespaces[1], def.files[1].namespaces[1].namespaces[0]);
            });
        });
        describe("getting the file and namespaces to a variable not in any file", function () {
            var result = def.getFileAndNamespacesToDefinition(createFunctions_1.createVariable({ name: "v" }));
            it("result should be null", function () {
                assert.equal(result, null);
            });
        });
    });
});

//# sourceMappingURL=globalDefinitionTests.js.map
